@@ -20,7 +20,7 @@
 #  respective owners and no grant or license is provided thereof.
 
 package PDL::IO::Touchstone;
-$VERSION = 1.008;
+$VERSION = 1.009;
 
 use 5.010;
 use strict;
@@ -28,6 +28,8 @@ use warnings;
 
 use Carp;
 
+use IO::Handle;
+
 use PDL;
 use PDL::LinearAlgebra;
 use PDL::Constants qw(PI);
@@ -40,6 +42,9 @@ BEGIN {
 	our @ISA = ( @ISA, qw(Exporter) );
 	our @EXPORT = qw/rsnp wsnp/;
 	our @EXPORT_OK = qw/
+		rsnp_fh
+		wsnp_fh
+
 		n_ports
 		m_interpolate
 		f_uniformity
@@ -89,16 +94,33 @@ sub rsnp
 {
 	my ($fn, $args) = @_;
 
+	$args = { ($args ? %$args : () ), filename => $fn };
+
+	open(my $in, $fn) or croak "$fn: $!";
+
+	my @ret = rsnp_fh($in, $args);
+
+	close($in) or carp "$fn: $!";
+
+	return @ret;
+}
+
+sub rsnp_fh
+{
+	my ($in, $args) = @_;
+
 	my $class;
 	my $comments = '';
 	my ($orig_funit, $param_type, $fmt, $R, $z0);
 	my $n_ports;
 
+	my $fn = $args->{filename} // '(unknown file)';
+
+	my $EOF_REGEX = $args->{EOF_REGEX};
+
 	# Try to enforce the number of ports based on the filename extension.
 	$n_ports = $1 if ($fn =~ /s(\d+)p/i);
 
-	open(my $in, $fn) or croak "$fn: $!";
-
 	my $n = 0;
 	my $line;
 
@@ -108,15 +130,27 @@ sub rsnp
 	my @comments;
 	my @cols;
 	my @f;
+
+	# input_line_number is a surprisingly slow call, so get it out of the loop:
+	$n = $in->input_line_number - 1;
+
 	while (defined($line = <$in>))
 	{
-		chomp($line);
 		$n++;
 
+		last if (defined($EOF_REGEX) && $line =~ /$EOF_REGEX/);
+
+		last if $line =~ /^\s*!\s*Noise Parameters/i;
+
+		chomp($line);
+
 		$line =~ s/^\s+|\s+$//g;
 
 		next if !length($line);
 
+		# Skip % lines as found in MDIF files:
+		next if $line =~ /^\s*%/;
+
 		# Strip leading space so split() will work properly.
 		$line =~ s/^\s+//;
 
@@ -329,11 +363,14 @@ sub wsnp_fh
 	# $out is in touchstone-formated order for each frequency with frequency as the first element:
 	# ie: [freq s11 21 s12 s11] < transposed for only for 2-port models.
 	
+	$f = $f->reshape($n_freqs);
+	my @freqs = $f->dog;
+
 	for (my $i = 0; $i < $n_freqs; $i++)
 	{
 		# matrix at frequency $i:
 		my $fm = $out->slice(":,$i");
-		my $freq = $f->slice("$i")->sclr;
+		my $freq = $freqs[$i];
 
 		# More than 2 ports are printed on multiple lines,
 		# at least one line for each port.
@@ -342,9 +379,14 @@ sub wsnp_fh
 			$fm = $fm->reshape($n_ports*2,$n_ports);
 		}
 
+		# First, print the frequency:
 		print $fd $freq;
 
 		# foreach matrix row:
+		# TODO: This loop is a hotspot, maybe someone can optimize it!
+		#       Using PDL::IO::Misc's `wcols` is probably a good idea
+		#       but formating it for Touchstone's weird column format
+		#       for n_ports >2 could be challenging.
 		foreach my $row ($fm->dog)
 		{
 			# No more than four data samples are allowed per line,
@@ -1342,6 +1384,21 @@ re-create the original touchstone file.
 
 =back
 
+=head2 C<rsnp_fh($fh, $options)> - Read touchstone file
+
+This is the same as C<rsnp> except that it takes a file handle instead of a
+filename.  Additionally, C<$options> accepts the following additional values:
+
+=over 4
+
+=item C<filename> - the original filename to facilitate more verbose error output.
+
+=item C<EOF_REGEX> - a regular expression that, when matched, will cause C<rsnp_fh> to stop reading data.
+
+This is used by L<PDL::IO::MDIF> when loading multiple touchstone files from a single MDIF file.
+
+=back
+
 =head2 C<wsnp($filename, $f, $m, $param_type, $z0, $comments, $fmt, $from_hz, $to_hz)>
 
 =head3 Arguments
@@ -1681,17 +1738,19 @@ For example, re-compose $T from the C<m_to_pos_vecs> example.
 
 =over 4
 
+=item L<PDL::IO::MDIF> - A L<PDL> IO module to load Measurement Data Interchange Format (*.mdf) files.
+
 =item L<RF::Component> - An object-oriented encapsulation of C<PDL::IO::Touchstone>.
 
 =item Touchstone specification: L<https://ibis.org/connector/touchstone_spec11.pdf>
 
 =item S-parameter matrix transform equations: L<http://qucs.sourceforge.net/tech/node98.html>
 
-=item Building MDF files from multiple S2P files: L<https://youtu.be/q1ixcb_mgeM>, L<https://github.com/KJ7NLL/mdf/>
+=item Building MDIF/MDF files from multiple S2P files: L<https://youtu.be/q1ixcb_mgeM>, L<https://github.com/KJ7NLL/mdf/>
 
 =item Optimizing amplifer impedance match circuits with MDF files: L<https://youtu.be/nx2jy7EHzxw>
 
-=item MDF file format: L<https://awrcorp.com/download/faq/english/docs/users_guide/data_file_formats.html#i489154>
+=item MDIF file format: L<https://awrcorp.com/download/faq/english/docs/users_guide/data_file_formats.html#i489154>
 
 =item "Conversions Between S, Z, Y, h, ABCD, and T Parameters which are Valid
 for Complex Source and Load Impedances" March 1994 IEEE Transactions on
